package com.laguerre.kan;

import android.opengl.Matrix;

public final class Matrix4 {
	public static final Matrix4 IDENTITY = getNewIdentity();

	/** Temporary memory for operations that need temporary matrix data. */
	private static final float[] sTemp = new float[32];

	private final float[] mat = new float[16];

	public Matrix4() {
		setToIdentity();
	}

	public Matrix4(Vector3 translation) {
		this(translation, Quaternion.IDENTITY);
	}

	public Matrix4(Vector3 translation, Quaternion rotation) {
		this(translation, rotation, Vector3.ONE);
	}

	public Matrix4(Vector3 translation, Quaternion rotation, Vector3 scale) {
		setToIdentity();
		scale(scale);
		rotate(rotation);
		translate(translation);
	}

	public Matrix4(Matrix4 mat) {
		setValues(mat);
	}

	public void setValues(Matrix4 mat) {
		System.arraycopy(mat.mat, 0, this.mat, 0, 16);
	}

	public void setValues(float[] mat4) {
		System.arraycopy(mat4, 0, this.mat, 0, 16);
	}

	public void setToIdentity() {
		Matrix.setIdentityM(this.mat, 0);
	}

	public void scale(Vector3 scale) {
		scale(scale.getX(), scale.getY(), scale.getZ());
	}

	public void scale(float sx, float sy, float sz) {
		Matrix.scaleM(this.mat, 0, sx, sy, sz);
	}

	public void rotate(Quaternion q) {
		final float[] angleAxis = new float[4];
		q.getAngleAxis(angleAxis);
		rotateByAngleAxis(angleAxis[0], angleAxis[1], angleAxis[2], angleAxis[3]);
	}

	public void rotateByAngleAxis(float angleDeg, float axeX, float axeY, float axeZ) {
		Matrix.setRotateM(this.mat, 0, angleDeg, axeX, axeY, axeZ);
	}

	public void translate(Vector3 v) {
		translate(v.getX(), v.getY(), v.getZ());
	}

	public void translate(float dx, float dy, float dz) {
		Matrix.translateM(this.mat, 0, dx, dy, dz);
	}

	public Vector3 extractScale(Vector3 result) {
		result.setValues(
				Matrix.length(this.mat[    0], this.mat[        1], this.mat[        2]),
				Matrix.length(this.mat[    4], this.mat[    4 + 1], this.mat[    4 + 2]),
				Matrix.length(this.mat[2 * 4], this.mat[2 * 4 + 1], this.mat[2 * 4 + 2]));
		return result;
	}
//
//	/**
//	 * Gets the scale contains in this {@link Matrix4} and its rotation part in
//	 * a matrix 3x3 column-major (as in openGL).
//	 * @param resultScale the extracted scale.
//	 * @param resultMat3 the extracted rotation in a matrix 3x3 column-major.
//	 */
//	public void extractScaleAndRotation(Vector3 resultScale, float[] resultMat3) {
//		extractScale(resultScale);
//		for (int i = 0; i < 3; i++) {
//			final float div = 1.0f / resultScale.getComponent(i);
//			for (int j = 0; j < 3; j++) {
//				resultMat3[i * 3 + j] = this.mat[i * 4 + j] * div;
//			}
//		}
//	}
//
//	public void extractScaleAndRotation(Vector3 resultScale, Quaternion resultRotation) {
//		final float[] mat3 = new float[9];
//		extractScaleAndRotation(resultScale, mat3);
//		resultRotation.setValuesFromMatrix3(mat3);
//	}
//
//	public float[] extractRotation(float[] resultMat3) {
//		final Vector3 dummyScale = new Vector3();
//		extractScaleAndRotation(dummyScale, resultMat3);
//		return resultMat3;
//	}
//
//	public Quaternion extractRotation(Quaternion result) {
//		final Vector3 dummyScale = new Vector3();
//		extractScaleAndRotation(dummyScale, result);
//		return result;
//	}

	public Vector3 extractTranslation(Vector3 result) {
		result.setValues(this.mat[3 * 4], this.mat[3 * 4 + 1], this.mat[3 * 4 + 2]);
		return result;
	}

	public float getComponent(int row, int column) {
		return this.mat[row * 4 + column];
	}

	public float getComponentAt(int pos) {
		return this.mat[pos];
	}

	public float[] getComponents(float[] result) {
		System.arraycopy(this.mat, 0, result, 0, 16);
		return result;
	}

	public float[] getSharedComponents() {
		return this.mat;
	}

	public Vector3 transform(Vector3 result, Vector3 p, float pW) {
		synchronized(sTemp) {
			p.getComponents(sTemp);
			sTemp[3] = pW;
			Matrix.multiplyMV(sTemp, 4, this.mat, 0, sTemp, 0);
			final float div = 1.0f / sTemp[7];
			result.setValues(sTemp[4] * div, sTemp[5] * div, sTemp[6] * div);
		}
		return result;
	}

	public Vector3 transformPoint(Vector3 result, Vector3 p) {
		return transform(result, p, 1.0f);
	}

	public Matrix4 getInverse(Matrix4 result) {
		Matrix.invertM(result.mat, 0, this.mat, 0);
		return result;
	}

	public Matrix4 getInverse() {
		final Matrix4 ret = new Matrix4();
		return getInverse(ret);
	}

	public static Matrix4 combine(Matrix4 result, Matrix4 lm, Matrix4 rm) {
		Matrix.multiplyMM(result.mat, 0, lm.mat, 0, rm.mat, 0);
		return result;
	}

	public static Matrix4 combine(Matrix4 lm, Matrix4 rm) {
		final Matrix4 ret = new Matrix4();
		return combine(ret, lm, rm);
	}

	public static Matrix4 getNewIdentity() {
		return new Matrix4();
	}
}
