package com.laguerre.kan;

import java.nio.ShortBuffer;

import android.opengl.GLES20;

public abstract class Renderer extends Component {
	/**
	 * Stores indices as an element buffer object and makes it possible to draw
	 * triangles associated with these indices.
	 */
	static class TriangleEBO {
		public static enum EBOPerfHint {
			STATIC_DRAW,
			STREAM_DRAW,
			DYNAMIC_DRAW
		}

		public final int eboId;
		private boolean beingNew = true;
		private int bufferSize = -1;
		private int nbIndices = -1;
		private EBOPerfHint perfHint;

		public TriangleEBO() {
			final int[] tmpId = { -1 };
			GLES20.glGenBuffers(1, tmpId, 0);
			this.eboId = tmpId[0];
		}

		public boolean isNew() {
			return this.beingNew;
		}

		public void createData(ShortBuffer triangleBuffer, EBOPerfHint perfHint) {
			this.beingNew = false;
			this.perfHint = perfHint;
			this.nbIndices = triangleBuffer.capacity();
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, this.eboId);
			int drawMode;
			switch (this.perfHint) {
			case STREAM_DRAW:
				drawMode = GLES20.GL_STREAM_DRAW;
				break;
			case DYNAMIC_DRAW:
				drawMode = GLES20.GL_DYNAMIC_DRAW;
				break;
			case STATIC_DRAW:
			default:
				drawMode = GLES20.GL_STATIC_DRAW;
				break;
			}
			this.bufferSize = triangleBuffer.capacity() * Short.SIZE / 8;
			GLES20.glBufferData(GLES20.GL_ELEMENT_ARRAY_BUFFER,
					this.bufferSize, triangleBuffer, drawMode);
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0);
		}

		public void updateData(ShortBuffer triangleBuffer) {
			if (this.perfHint == EBOPerfHint.STATIC_DRAW) {
				return;
			}
			final int newByteSize = triangleBuffer.capacity() * Short.SIZE / 8;
			if (newByteSize != this.bufferSize) {
				createData(triangleBuffer, this.perfHint);
			} else {
				GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, this.eboId);
				GLES20.glBufferSubData(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0,
						this.bufferSize, triangleBuffer);
				GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0);
			}
		}

		public void drawElements() {
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, this.eboId);
			AndroidGL20Compact.getInstance().glDrawElements(GLES20.GL_TRIANGLES,
					this.nbIndices, GLES20.GL_UNSIGNED_SHORT, 0);
//			GLES20.glDrawElements(GLES20.GL_TRIANGLES,
//					this.nbIndices, GLES20.GL_UNSIGNED_SHORT, 0);
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0);
		}

		public void freeResources() {
			final int[] tmpId = { this.eboId };
			GLES20.glDeleteBuffers(1, tmpId, 0);
		}
	}

	protected final Bounds worldBounds = new Bounds(Vector3.ZERO, Vector3.ZERO);

	public abstract void setMaterial(Material mat, int subMeshId);

	public abstract Material getSharedMaterial(int subMeshId);

	public abstract int findSubMeshIdOfMaterial(String materialName);

	public abstract Material findMaterial(String materialName);

	public abstract void setToFirstPass();

	public abstract int getCurrentPass();

	public abstract boolean renderOnePass(Camera camera);

	public abstract boolean stillNeedsPass();

	public final Bounds getSharedWorldBounds() {
		return this.worldBounds;
	}

	abstract void updateWorldBounds();

	/**
	 * Stores the 8 vertices of an AABB. Used by
	 * <code>updateWorldBounds(Bounds localBounds)</code> only to avoid many
	 * memory allocations.
	 */
	private static final Vector3[] points_worldBounds = {
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec()
	};
	/**
	 * Stores the minimum values from the 8 vertex coordinates of an AABB.
	 * Used by <code>updateWorldBounds(Bounds localBounds)</code> only to avoid
	 * many memory allocations.
	 */
	private static final Vector3 min_worldBounds = Vector3.getNewZeroVec();
	/**
	 * Stores the maximum values from the 8 vertex coordinates of an AABB.
	 * Used by <code>updateWorldBounds(Bounds localBounds)</code> only to avoid
	 * many memory allocations.
	 */
	private static final Vector3 max_worldBounds = Vector3.getNewZeroVec();

	protected final void updateWorldBounds(Bounds localBounds) {
		synchronized (points_worldBounds) {
			localBounds.getHeightVertices(points_worldBounds);
			final Transform t = getSharedTransform();
			for (int i = 0; i < points_worldBounds.length; i++) {
				t.transformWorldPoint(
						points_worldBounds[i], points_worldBounds[i]);
			}

			float minX = Float.MAX_VALUE, minY = Float.MAX_VALUE, minZ = Float.MAX_VALUE;
			float maxX = Float.MIN_VALUE, maxY = Float.MIN_VALUE, maxZ = Float.MIN_VALUE;
			for (int i = 0; i < points_worldBounds.length; i++) {
				final float x = points_worldBounds[i].getX();
				final float y = points_worldBounds[i].getY();
				final float z = points_worldBounds[i].getZ();
				if (x < minX) {
					minX = x;
				}
				if (y < minY) {
					minY = y;
				}
				if (z < minZ) {
					minZ = z;
				}
				if (x > maxX) {
					maxX = x;
				}
				if (y > maxY) {
					maxY = y;
				}
				if (z > maxZ) {
					maxZ = z;
				}
			}
			min_worldBounds.setValues(minX, minY, minZ);
			max_worldBounds.setValues(maxX, maxY, maxZ);
			this.worldBounds.setFromMinAndMax(
					min_worldBounds, max_worldBounds);
		}
	}
}
